#   pyRCV2: Preferential vote counting
#   Copyright © 2020  Lee Yingtong Li (RunasSudo)
#
#   This program is free software: you can redistribute it and/or modify
#   it under the terms of the GNU Affero General Public License as published by
#   the Free Software Foundation, either version 3 of the License, or
#   (at your option) any later version.
#
#   This program is distributed in the hope that it will be useful,
#   but WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#   GNU Affero General Public License for more details.
#
#   You should have received a copy of the GNU Affero General Public License
#   along with this program.  If not, see <https://www.gnu.org/licenses/>.

__pragma__ = lambda x: None

from pyRCV2.method.base_stv import UIGSTVCounter
from pyRCV2.model import CandidateState, CountCard
from pyRCV2.numbers import Num, Rational
from pyRCV2.safedict import SafeDict

class ParcelledEGCountCard(CountCard):
	def __init__(self, *args):
		CountCard.__init__(self, *args)
		del self.ballots
		self.parcels = []
	
	#@property
	#def ballots(self):
	#	raise Exception('Attempted to access ballots attribute of ParcelledUIGCountCard')
	
	def clone(self):
		"""Overrides CountCard.clone"""
		result = ParcelledUIGCountCard()
		result.orig_votes = self.orig_votes
		result.transfers = self.transfers
		result.parcels = [[(b[0].clone(), b[1]) for b in p] for p in self.parcels]
		result.state = self.state
		result.order_elected = self.order_elected
		return result

class ParcelledEGSTVCounter(UIGSTVCounter):
	"""
	Exclusive Gregory (parcel by parcel) STV implementation
	"""
	
	def __init__(self, *args):
		UIGSTVCounter.__init__(self, *args)
		self.candidates = SafeDict([(c, ParcelledEGCountCard()) for c in self.election.candidates])
	
	def distribute_first_preferences(self):
		"""Overrides UIGSTVCounter.distribute_first_preferences"""
		
		for ballot in self.election.ballots:
			__pragma__('opov')
			candidate = next((c for c in ballot.preferences if self.candidates[c].state == CandidateState.HOPEFUL), None)
			
			if candidate is not None:
				self.candidates[candidate].transfers += ballot.value
				if len(self.candidates[candidate].parcels) == 0:
					self.candidates[candidate].parcels.append([(ballot, self.cls_ballot_value(ballot.value))])
				else:
					self.candidates[candidate].parcels[0].append((ballot, self.cls_ballot_value(ballot.value)))
			else:
				self.exhausted.transfers += ballot.value
				self.exhausted.ballots.append((ballot, self.cls_ballot_value(ballot.value)))
			__pragma__('noopov')
	
	def do_surplus(self, candidate_surplus, count_card, surplus):
		"""Overrides UIGSTVCounter.do_surplus"""
		
		next_preferences = SafeDict([(c, []) for c, cc in self.candidates.items()])
		total_ballots = Num('0')
		total_votes = Rational('0')
		
		parcel = count_card.parcels[len(count_card.parcels)-1] # Last bundle only
		
		__pragma__('opov')
		
		# Count next preferences
		for ballot, ballot_value in parcel:
			candidate = next((c for c in ballot.preferences if self.candidates[c].state == CandidateState.HOPEFUL), None)
			
			if candidate is not None:
				# Examine only transferable ballot papers
				total_ballots += ballot.value
				total_votes += ballot_value
				next_preferences[candidate].append((ballot, ballot_value))
		
		# Make transfers
		for candidate, cand_ballots in next_preferences.items():
			if len(cand_ballots) == 0:
				continue
			
			new_parcel = []
			self.candidates[candidate].parcels.append(new_parcel)
			
			if total_votes.to_num() > surplus:
				# Reweight according to transfer value
				num_ballots = sum((b.value for b, bv in cand_ballots), Num('0'))
				self.candidates[candidate].transfers += (num_ballots * surplus) / total_ballots # Multiply first to avoid rounding errors
				
				for ballot, ballot_value in cand_ballots:
					new_value = (ballot.value.to_rational() * surplus.to_rational()) / total_ballots.to_rational()
					new_parcel.append((ballot, new_value))
			else:
				# Do not reweight
				num_votes = sum((bv for b, bv in cand_ballots), Rational('0'))
				self.candidates[candidate].transfers += num_votes.to_num()
				
				for ballot, ballot_value in cand_ballots:
					new_parcel.append((ballot, ballot_value))
		
		# Credit exhausted votes (only if not reweighting)
		if total_votes.to_num() <= surplus:
			self.exhausted.transfers += surplus - total_votes.to_num()
		
		__pragma__('noopov')
		
		return True
	
	def do_exclusion(self, candidate_excluded, count_card):
		"""Overrides UIGSTVCounter.do_exclusion"""
		
		next_preferences = SafeDict([(c, []) for c, cc in self.candidates.items()])
		next_exhausted = []
		
		parcel = count_card.parcels[0]
		count_card.parcels.remove(parcel)
		
		# Count next preferences
		for ballot, ballot_value in parcel:
			__pragma__('opov')
			candidate = next((c for c in ballot.preferences if self.candidates[c].state == CandidateState.HOPEFUL), None)
			__pragma__('noopov')
			
			if candidate is not None:
				__pragma__('opov')
				next_preferences[candidate].append((ballot, ballot_value))
				__pragma__('noopov')
			else:
				next_exhausted.append((ballot, ballot_value))
		
		# Make transfers
		for candidate, cand_ballots in next_preferences.items():
			if len(cand_ballots) == 0:
				continue
			
			new_parcel = []
			__pragma__('opov')
			self.candidates[candidate].parcels.append(new_parcel)
			__pragma__('noopov')
			
			num_votes = sum((bv for b, bv in cand_ballots), Rational('0'))
			__pragma__('opov')
			self.candidates[candidate].transfers += num_votes.to_num()
			count_card.transfers -= num_votes.to_num()
			__pragma__('noopov')
			
			for ballot, ballot_value in cand_ballots:
				new_parcel.append((ballot, ballot_value))
		
		# Credit exhausted votes
		num_votes = sum((bv for b, bv in next_exhausted), Rational('0'))
		__pragma__('opov')
		self.exhausted.transfers += num_votes.to_num()
		count_card.transfers -= num_votes.to_num()
		__pragma__('noopov')
		
		for ballot, ballot_value in next_exhausted:
			self.exhausted.ballots.append((ballot, ballot_value))
		
		if len(count_card.parcels) == 0:
			count_card.state = CandidateState.EXCLUDED
